/**********************************************************************
 * perditiondb_posix_regex.c                              December 1999
 * Horms                                             horms@vergenet.net
 *
 * Access a posix_regex map
 *
 * Released under the terms of the GNU GPL
 **********************************************************************/

#include "perditiondb_posix_regex.h"

static dynamic_array_t *regex_a;


/**********************************************************************
 * dbserver_init
 * Read the server information for a given key from the posix_regex map
 * pre: options_str: options string to use
 *                   Specifies the name of the regex map to open
 *                   If NULL then PERDITIONDB_POSIX_REGEX_MAPNAME
 *                   is used
 * post: regex map is opened, regex are compiled and stored with
 *       their server in the static dynamic_array_t *regex_a
 * return:  0 on sucess
 *         -1 on file access error
 *         -2 if key cannot be found in map
 *         -3 on other error
 **********************************************************************/

int dbserver_init(char *options_str){
  key_value_t *kv=NULL;
  regex_t *preg=NULL;
  FILE *stream=NULL;
  char *line=NULL;
  char *line_cur;
  char *line_end;
  char *regex;
  char *server_port_str;
  int blank;

  extern int errno;

  regex_a=NULL;
 
  if((stream=fopen(
    (options_str==NULL)?PERDITIONDB_POSIX_REGEX_MAPNAME:options_str, "r")
  )==NULL){
    fprintf(
      stderr,
      "Could not open %s: %s\n",
      (options_str==NULL)?PERDITIONDB_POSIX_REGEX_MAPNAME:options_str,
      strerror(errno)
    );
    return(-1);
  }

  if((line=(char *)malloc(PERDITIONDB_POSIX_REGEX_MAX_LINE_LENGTH))==NULL){
    goto leave_3;
  }
  if((regex_a=dynamic_array_create(0))==NULL){ goto leave_3; }
  if((kv=key_value_create())==NULL){ goto leave_3; }

  line_end=line+PERDITIONDB_POSIX_REGEX_MAX_LINE_LENGTH;
  while(fgets(line, PERDITIONDB_POSIX_REGEX_MAX_LINE_LENGTH, stream)!=NULL){
    blank=1;
    regex=line;
    server_port_str=NULL;
    *line_end='\0';
    for(line_cur=line;line_cur<line_end;line_cur++){
      if(*line_cur=='\0'||*line_cur=='\n'||*line_cur=='\r' || *line_cur=='#'){
        *line_cur='\0';
        break;
      }
      if(blank && *line_cur!=' ' && *line_cur!='\t'){
        blank=0;
      }
      if(blank){
        (server_port_str==NULL)?regex++:server_port_str++;
      }
      if(
        server_port_str==NULL && 
        *line_cur==PERDITIONDB_POSIX_REGEX_MAX_FIELD_DELIMITER
      ){
        *line_cur='\0';
        server_port_str=line_cur+1;
        blank=1;
        continue;
      }
    }
    if(blank||(*regex=='\0'||server_port_str==NULL||*server_port_str=='\0')){
      continue;
    }
    if((preg=(regex_t *)malloc(sizeof(regex_t)))==NULL){ goto leave_3; }
    if(regcomp(preg, regex, REG_EXTENDED|REG_NOSUB|REG_NEWLINE)){
      goto leave_3;
    }
    if((kv=key_value_assign(
      kv, 
      (void *)preg, 
      DESTROY_REGEX, 
      NULL,
      (void *)server_port_str,
      DESTROY_STR,
      DUP_STR
    ))==NULL){ 
      goto leave_3; 
    }
    if(dynamic_array_add_element(regex_a,kv,DESTROY_KV,DUP_KV)==NULL){
      goto leave_3;
    }
  }

  fclose(stream);
  free(line);
  return(0);
  free(kv);

  leave_3:
  if(stream!=NULL){ fclose(stream); }
  if(line!=NULL){ free(line); }
  if(regex_a!=NULL){ dynamic_array_destroy(regex_a, DESTROY_REGEX); }
  if(preg!=NULL){ destroy_regex(preg); }
  if(kv!=NULL){ key_value_destroy(kv); }
  return(-3);
} 


/**********************************************************************
 * dbserver_get
 * Find the server for a given user
 * pre: key_str: user to find server for
 *      options_str: options string
 *                   ignored
 *      str_return:  Value is returned here
 *      len_return:  Length of value is returned here
 * post: The str_key is looked up and the corresponding value is 
 *       returned in str_return and len_return.
 * return:  0 on sucess
 *         -1 on file access error
 *         -2 if key cannot be found in map
 *         -3 on other error
 * Note: The string returned in str_return should be of the 
 * form <servername>[:<port>].
 * E.g.: localhost:110
 *       localhost 
 **********************************************************************/

int dbserver_get(
  char *key_str, 
  char *options_str,
  char **str_return, 
  int  *len_return
){
  int i;
  int status;
  key_value_t *kv;
  char *server_port_str;

  for(i=dynamic_array_get_count(regex_a)-1;i>=0;i--){
    kv=dynamic_array_get_element(regex_a,i);
    status=regexec((regex_t *)key_value_get_key(kv), key_str, 0, NULL, 0);
    if(!status){
      server_port_str=(char *)key_value_get_value(kv);
      *str_return=strdup(server_port_str);
      *len_return=strlen(server_port_str);
      return(0);
    }
  }

  return(-2);
} 


/**********************************************************************
 * dbserver_fini
 * Free memory structures associated with the regex map
 * pre: none
 * post: static dynamic_array_t *regex_t is freed
 * return:  0 on sucess
 *         -1 on file access error
 *         -2 if key cannot be found in map
 *         -3 on other error
 **********************************************************************/

int dbserver_fini(void){
  dynamic_array_destroy(regex_a, DESTROY_KV);
  return(0);
}


/**********************************************************************
 * dummy_dup_regex
 * Dummy funtion to duplicate a regex.
 * Does nothing, returns preg, not a copy
 * pre: preg: regex to copy
 * post: nothing
 * return: preg
 **********************************************************************/

static regex_t *dummy_dup_regex(regex_t *preg){
  return(preg);
}


/**********************************************************************
 * destroy_regex
 * Free up the memory associated with a regex
 * pre: preg: regex to free
 * post: regex structure and contents is freed
 * return: nothing
 **********************************************************************/

static void destroy_regex(regex_t *preg){
  regfree(preg);
  free(preg);
}

